import argparse
import json
from pathlib import Path
from typing import Dict

import torch
import yaml
from torchmetrics.classification import MulticlassJaccardIndex  # type: ignore
from tqdm import tqdm

from phenobench.evaluation.auxiliary.common import (
    get_png_files_in_dir,
    load_file_as_int_tensor,
    load_file_as_tensor,
)
from phenobench.evaluation.auxiliary.convert import convert_partial_semantics

torch.set_printoptions(precision=4, sci_mode=False)


def parse_args() -> Dict:
  parser = argparse.ArgumentParser()
  parser.add_argument('--phenobench_dir', required=True, type=Path, help='Path to the root directory of PhenoBench.')
  parser.add_argument('--prediction_dir', required=True, type=Path, help='Path to prediction directory.')
  parser.add_argument('--export', required=False, type=Path, help='Path to export directory.')
  parser.add_argument('--split',
                      choices=['train', 'val', 'test'],
                      required=True,
                      type=str,
                      help='Specify which split to use for evaluation.')
  parser.add_argument('--mapping', required=False, type=Path, help='Path to mapping.json file (only for test-split)')

  # Required directory structure
  # ├── PhenoBench
  # |   └── <split>
  # │       └── semantics
  # └── prediction
  #     └── semantics

  args = vars(parser.parse_known_args()[0])
  if "export" in args: args['export'].mkdir(parents=True, exist_ok=True)

  if args['split'] == 'test':
    if (not args['mapping']):
      raise TypeError(f"For the evaluation on the test-split you need to specify the '--mapping' arguments.")

  return args


def evaluate_semantics(args: Dict[str, str]) -> Dict[str, float]:
  """ Compute semantic segmentation scores: IoU for "crop", "weed", "soil" and "mIoU".

  Args:
      args (Dict): command line arguments specifying the required paths

  Returns:
      Dict: metrics for classes "soil", "crop", "weed"
  """

  # ------- Get all ground truth and prediction files -------
  gt_fnames = get_png_files_in_dir(args['phenobench_dir'] / args['split'] / 'semantics')
  pred_fnames = get_png_files_in_dir(args['prediction_dir'] / 'semantics')

  # Load the mapping from randomized to original fnames
  if args['split'] == 'test':
    with open(args['mapping'], 'r') as istream:
      fname_mapping = json.load(istream)
    fname_mapping_reverse = {val: key for key, val in fname_mapping.items()}

    pred_original_fnames = []
    for pred_fname in pred_fnames:
      pred_original_fnames.append(fname_mapping[pred_fname])
    gt_fnames.sort()
    pred_original_fnames.sort()
  else:
    pred_original_fnames = pred_fnames
    fname_mapping_reverse = {fname: fname for fname in pred_fnames}

  # ------- Setup evaluator -------
  evaluator = MulticlassJaccardIndex(num_classes=3, average=None)
  if args['split'] == 'test':
    evaluator_2020 = MulticlassJaccardIndex(num_classes=3, average=None)
    evaluator_2021 = MulticlassJaccardIndex(num_classes=3, average=None)

  n_total = len(gt_fnames)
  for gt_fname, pred_fname in tqdm(zip(gt_fnames, pred_original_fnames), total=n_total):
    assert gt_fname == pred_fname

    semantics_gt = convert_partial_semantics(
        load_file_as_tensor(args['phenobench_dir'] / args['split'] / 'semantics' / gt_fname)).squeeze()
    semantics_pred = convert_partial_semantics(
        load_file_as_int_tensor(args['prediction_dir'] / 'semantics' / fname_mapping_reverse[pred_fname])).squeeze()

    if args['split'] == 'test':
      if '2021' not in gt_fname:
        evaluator_2020.update(semantics_pred, semantics_gt)
      if '2021' in gt_fname:
        evaluator_2021.update(semantics_pred, semantics_gt)

    evaluator.update(semantics_pred, semantics_gt)

  metrics = evaluator.compute() * 100.0  # tensor of shape [3]

  eval_results = {}
  if args['split'] == 'test':
    ext = "_all"
  else:
    ext = ""
  eval_results[f"soil{ext}"] = round(float(metrics[0]), 2)
  eval_results[f"crop{ext}"] = round(float(metrics[1]), 2)
  eval_results[f"weed{ext}"] = round(float(metrics[2]), 2)
  eval_results[f"mIoU{ext}"] = round(float(metrics.mean()), 2)

  if "export" in args:
    fpath_out_all = args['export'] / 'all'
    fpath_out_all.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_all / 'eval_semantics.yaml', 'w') as ostream:
      yaml.dump(eval_results, ostream)

  if args['split'] == 'test':
    # 2020
    eval_results_2020 = {}
    metrics_2020 = evaluator_2020.compute() * 100.0
    ext = "_2020"
    eval_results_2020[f"soil{ext}"] = round(float(metrics_2020[0]), 2)
    eval_results_2020[f"crop{ext}"] = round(float(metrics_2020[1]), 2)
    eval_results_2020[f"weed{ext}"] = round(float(metrics_2020[2]), 2)
    eval_results_2020[f"mIoU{ext}"] = round(float(metrics_2020.mean()), 2)

    fpath_out_2020 = args['export'] / '2020'
    fpath_out_2020.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_2020 / 'eval_semantics.yaml', 'w') as ostream:
      yaml.dump(eval_results_2020, ostream)

    # 2021
    eval_results_2021 = {}
    metrics_2021 = evaluator_2021.compute() * 100.0
    ext = "_2021"
    eval_results_2021[f"soil{ext}"] = round(float(metrics_2021[0]), 2)
    eval_results_2021[f"crop{ext}"] = round(float(metrics_2021[1]), 2)
    eval_results_2021[f"weed{ext}"] = round(float(metrics_2021[2]), 2)
    eval_results_2021[f"mIoU{ext}"] = round(float(metrics_2021.mean()), 2)

    if "export" in args:
      fpath_out_2021 = args['export'] / '2021'
      fpath_out_2021.mkdir(parents=True, exist_ok=True)
      with open(fpath_out_2021 / 'eval_semantics.yaml', 'w') as ostream:
        yaml.dump(eval_results_2021, ostream)

  return eval_results


if __name__ == '__main__':
  args = parse_args()
  evaluate_semantics(args)



import argparse
import json
from pathlib import Path
from typing import Dict

import torch
import yaml
from torchmetrics.detection.mean_ap import MeanAveragePrecision  # type: ignore
from tqdm import tqdm

from phenobench.evaluation.auxiliary.common import (
    get_png_files_in_dir,
    get_txt_files_in_dir,
    load_file_as_tensor,
    load_yolo_txt_file,
)
from phenobench.evaluation.auxiliary.convert import (
    cvt_gt_to_bbox_map,
    cvt_yolo_to_bbox_map,
)
from phenobench.evaluation.auxiliary.filter import filter_partials_boxes


def parse_args() -> Dict:
  parser = argparse.ArgumentParser()
  parser.add_argument('--phenobench_dir', required=True, type=Path, help='Path to ground truth directory.')
  parser.add_argument('--prediction_dir', required=True, type=Path, help='Path to prediction directory.')
  parser.add_argument('--export', required=True, type=Path, help='Path to export directory.')
  parser.add_argument('--split', choices=['train', 'val', 'test'], required=True, type=str, help='Specify which split to use for evaluation.')
  
  parser.add_argument('--mapping', required=False, type=Path, help='Path to mapping.json file (only for test-split)')
  # Required directory structure
  # ├── ground-truth
  # │   ├── leaf_instances
  # │   ├── leaf_visibility
  # └── prediction
  #     └── leaf_bboxes

  args = vars(parser.parse_args())

  args['export'].mkdir(parents=True, exist_ok=True)

  if args['split'] == 'test':
    if (not args['mapping']):
      raise TypeError(f"For the evaluation on the test-split you need to specify the '--date' and '--mapping' arguments.")

  return args


def evaluate_leaf_detection(args:Dict[str,str]) -> Dict[str,str]:

  # ------- Ground Truth -------
  gt_instance_fnames = get_png_files_in_dir(args['phenobench_dir']  / args["split"] / 'leaf_instances')
  gt_visibility_fnames = get_png_files_in_dir(args['phenobench_dir']  / args["split"] / 'leaf_visibility')

  # ------- Predictions -------
  pred_bboxes_fnames = get_txt_files_in_dir(args['prediction_dir'] / 'leaf_bboxes')

  # ------- Load the mapping from randomized to original fnames -------
  if args['split'] == 'test':
    with open(args['mapping'], 'r') as istream:
      fname_mapping = json.load(istream)
    fname_mapping_reverse = {val: key for key, val in fname_mapping.items()}

    pred_bboxes_original_fnames = []
    for pred_fname in pred_bboxes_fnames:
      pred_bboxes_original_fnames.append(fname_mapping[pred_fname])

    gt_instance_fnames.sort()
    gt_visibility_fnames.sort()
    pred_bboxes_original_fnames.sort()
  else:
    pred_bboxes_original_fnames = pred_bboxes_fnames
    fname_mapping_reverse = {fname:fname for fname in pred_bboxes_fnames}

  # ------- Setup evaluator -------
  evaluator = MeanAveragePrecision(box_format='cxcywh')
  if args['split'] == 'test':
    evaluator_2020 = MeanAveragePrecision(box_format='cxcywh')
    evaluator_2021 = MeanAveragePrecision(box_format='cxcywh')

  for gt_instance_fname, gt_visibility_fname, pred_bboxes_fnames in tqdm(zip(gt_instance_fnames, gt_visibility_fnames, pred_bboxes_original_fnames),total=len(gt_instance_fnames)):
    assert gt_instance_fname.split('.')[0] == gt_visibility_fname.split('.')[0] == pred_bboxes_fnames.split('.')[0]

    gt_instance_map = load_file_as_tensor(args['phenobench_dir']  / args["split"] / 'leaf_instances' / gt_instance_fname).squeeze()  # [H x W]
    gt_visibility = load_file_as_tensor(args['phenobench_dir']  / args["split"] / 'leaf_visibility' / gt_visibility_fname).squeeze()  # [H x W]
    gt_semantics = (gt_instance_map > 0).type(torch.uint8)  # derive semantics from instance map
    targets = cvt_gt_to_bbox_map(gt_instance_map, gt_semantics, gt_visibility)

    yolo_pred = load_yolo_txt_file(args['prediction_dir'] / 'leaf_bboxes' / fname_mapping_reverse[pred_bboxes_fnames])
    preds = cvt_yolo_to_bbox_map(yolo_pred)

    filter_partials_boxes(preds[0], targets[0])

    if args['split'] == 'test':
      if '2021' not in gt_instance_fname:
        evaluator_2020.update(preds, targets)
      if '2021' in gt_instance_fname:
        evaluator_2021.update(preds, targets)

    evaluator.update(preds, targets)

  metrics = evaluator.compute()
  eval_results = {}
  if args['split'] == 'test':
    ext = "_all"
  else:
    ext = ""
  eval_results[f'mAP{ext}'] = round(float(metrics['map'] * 100),2)
  eval_results[f'mAP_50{ext}'] = round(float(metrics['map_50'] * 100), 2)
  eval_results[f'mAP_75{ext}'] = round(float(metrics['map_75'] * 100), 2)
  
  if "export" in args:
    fpath_out_all = args['export'] / 'all'
    fpath_out_all.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_all / 'eval_leaf_bboxes.yaml', 'w') as ostream:
      yaml.dump(eval_results, ostream)

  if args['split'] == 'test':
    metrics_2020 = evaluator_2020.compute()
    eval_results_2020 = {}
    ext = "_2020"
    eval_results_2020[f'mAP{ext}'] = round(float(metrics_2020['map'] * 100), 2)
    eval_results_2020[f'mAP_50{ext}'] = round(float(metrics_2020['map_50'] * 100), 2)
    eval_results_2020[f'mAP_75{ext}'] = round(float(metrics_2020['map_75'] * 100), 2)
    
    fpath_out_2020 = args['export'] / '2020'
    fpath_out_2020.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_2020 / 'eval_leaf_bboxes.yaml', 'w') as ostream:
      yaml.dump(eval_results_2020, ostream)

    metrics_2021 = evaluator_2021.compute()
    eval_results_2021 = {}
    ext = "_2021"
    eval_results_2021[f'mAP{ext}'] = round(float(metrics_2021['map'] * 100), 2)
    eval_results_2021[f'mAP_50{ext}'] = round(float(metrics_2021['map_50'] * 100), 2)
    eval_results_2021[f'mAP_75{ext}'] = round(float(metrics_2021['map_75'] * 100), 2)
    
    if "export" in args:
      fpath_out_2021 = args['export'] / '2021'
      fpath_out_2021.mkdir(parents=True, exist_ok=True)
      with open(fpath_out_2021 / 'eval_leaf_bboxes.yaml', 'w') as ostream:
        yaml.dump(eval_results_2021, ostream)

  return eval_results

if __name__ == '__main__':
  args = parse_args()
  evaluate_leaf_detection(args)



import os
from typing import Dict, List

import numpy as np
from PIL import Image


class PhenoBench:
  """ `PhenoBench <https://www.phenobench.org/>`_ Dataset.

    This is a very basic dataloader with some utility function that can be easily integrated into a PyTorch, 
    PyTorch Lightning, etc. data loader. We kept the requirements low profile to ensure extensibility.
  
    Args:
      root (str): Root directory of the PhenoBench dataset, where the folders `train`, `val`, and `test` data is location.
      split (str, optional): The image split to use, `train`, `test`, or `val`.
      target_type (List[str], optional): Types of targets to use and provide with each sample, i.e., `semantics`, `plant_instances`, `leaf_instances`, `plant_bboxes`, `leaf_bboxes`, `plant_visibility`, and/or `leaf_visibility`.
      ignore_partial (bool, optional): Ignore partial annotations? if `True` all partial plants, i.e., will be mask using the `ignore_mask` in `semantics` and mapped to `0` in the instance masks. Default: `False`.
      make_unique_ids (bool, optional): Make the instance ids of weeds and crops unique. Default: True.
      ignore_mask(int, optional): ignore_index. Default: 255.
  """
  def __init__(self,
               root: str,
               split: str = 'train',
               target_types: List[str] = ['semantics'],
               ignore_partial: bool = False,
               make_unique_ids: bool = True,
               ignore_mask: int = 255) -> None:

    # sanity checks:
    root = os.path.expanduser(root)
    assert os.path.exists(root), f"The path to the dataset does not exist: `{root}`."
    assert split in ['train', 'val', 'test']
    for target_type in target_types:
      assert target_type in [
          'semantics', 'plant_instances', 'leaf_instances', 'plant_bboxes', 'leaf_bboxes', 'plant_visibility',
          'leaf_visibility'
      ]

    self.root = root
    self.split = split
    self.target_types = target_types
    self.ignore_partial = ignore_partial
    self.make_unique_ids = make_unique_ids
    self.ignore_mask = ignore_mask

    self.filenames = sorted(os.listdir(os.path.join(self.root, split, "images")))

  def __getitem__(self, index: int) -> Dict[str, np.array]:
    """
    Args:
        index (int): index of the image & annotation

    Returns:
        Dict[str, Any]: returns a sample with the specified target_types.

        A data point is comprised of the following information (at least):
        ```
        {
          "image_name": "
          "image": PIL Image with size W x H.
          ...
          <target>: np.array of shape W x H or a dictionary with the bounding boxes.
        }
        ```
    """
    sample = {}

    sample["image_name"] = self.filenames[index]
    sample["image"] = Image.open(os.path.join(self.root, self.split, "images", self.filenames[index])).convert("RGB")

    if self.split in ["train", "val"]:

      for target in self.target_types:
        if target in ["semantics", "plant_instances", "leaf_instances", "plant_visibility", "leaf_visibility"]:
          sample[target] = np.array(Image.open(os.path.join(self.root, self.split, target, self.filenames[index])))

      if self.ignore_partial:
        semantics = np.array(Image.open(os.path.join(self.root, self.split, "semantics", self.filenames[index])))
        partial_crops = semantics == 3
        partial_weeds = semantics == 4

        if "semantics" in self.target_types:
          sample["semantics"][partial_crops] = self.ignore_mask
          sample["semantics"][partial_weeds] = self.ignore_mask

        for target in self.target_types:
          if target in ["plant_instances", "leaf_instances"]:
            sample[target][partial_crops] = 0
            sample[target][partial_weeds] = 0
          ## FIXME: do something for the bounding boxes.
          if target in ["plant_bboxes", "leaf_bboxes"]:
            pass

      else:
        if "semantics" in self.target_types:
          # remap partial_crop to crop and partial_weed to weed:
          sample["semantics"][sample["semantics"] == 3] = 1
          sample["semantics"][sample["semantics"] == 4] = 2

        if "plant_bboxes" in self.target_types:
          semantics = np.array(Image.open(os.path.join(self.root, self.split, "semantics", self.filenames[index])))
          plant_instances = np.array(
              Image.open(os.path.join(self.root, self.split, "plant_instances", self.filenames[index])))
          leaf_instances = np.array(
              Image.open(os.path.join(self.root, self.split, "leaf_instances", self.filenames[index])))

          sample["plant_bboxes"] = []
          for label in [1, 2]:
            for plant_id in np.unique(plant_instances[semantics == label]):
              ys, xs = np.where((plant_instances == plant_id) & (semantics == label))

              width, height = np.max(xs) - np.min(xs), np.max(ys) - np.min(ys)
              center = (np.min(xs) + width // 2, np.min(ys) + height // 2)
              sample["plant_bboxes"].append({
                  "label": label,
                  "corner": (np.min(xs), np.min(ys)),
                  "center": center,
                  "width": width,
                  "height": height
              })

        if "leaf_bboxes" in self.target_types:
          semantics = np.array(Image.open(os.path.join(self.root, self.split, "semantics", self.filenames[index])))
          leaf_instances = np.array(
              Image.open(os.path.join(self.root, self.split, "leaf_instances", self.filenames[index])))

          sample["leaf_bboxes"] = []
          for leaf_id in np.unique(leaf_instances[leaf_instances > 0]):
            ys, xs = np.where((leaf_instances == leaf_id))

            width, height = np.max(xs) - np.min(xs), np.max(ys) - np.min(ys)
            center = (np.min(xs) + width // 2, np.min(ys) + height // 2)
            sample["leaf_bboxes"].append({
                "label": 1,
                "corner": (np.min(xs), np.min(ys)),
                "center": center,
                "width": width,
                "height": height
            })

    # remapping the instance ids to consecutive ids:
    if self.make_unique_ids:

      def replace(array: np.array, values, replacements):
        temp_array = array.copy()

        for v, r in zip(values, replacements):
          temp_array[array == v] = r

        array = temp_array

      if "plant_instances" in self.target_types:
        semantics = np.array(Image.open(os.path.join(self.root, self.split, "semantics", self.filenames[index])))

        crop_ids = np.unique(sample["plant_instances"][semantics == 1])
        weed_ids = np.unique(sample["plant_instances"][semantics == 2])

        N, M = len(crop_ids), len(weed_ids)

        replace(sample["plant_instances"][semantics == 1], crop_ids, np.arange(1, N + 1))
        replace(sample["plant_instances"][semantics == 2], weed_ids, np.arange(N + 1, N + M + 1))

      if "leaf_instances" in self.target_types:
        leaf_ids = np.unique(sample["leaf_instances"][sample["leaf_instances"] > 0])

        replace(sample["leaf_instances"], leaf_ids, np.arange(1, len(leaf_ids) + 1))

    return sample

  def __len__(self):
    return len(self.filenames)



import argparse
import json
from pathlib import Path
from typing import Dict

import yaml
from torchmetrics.detection.mean_ap import MeanAveragePrecision  # type: ignore
from tqdm import tqdm

from phenobench.evaluation.auxiliary.common import (
    get_png_files_in_dir,
    get_txt_files_in_dir,
    load_file_as_tensor,
    load_yolo_txt_file,
)
from phenobench.evaluation.auxiliary.convert import (
    cvt_gt_to_bbox_map,
    cvt_yolo_to_bbox_map,
)
from phenobench.evaluation.auxiliary.filter import filter_partials_boxes

IMG_WIDTH = 1024
IMG_HEIGHT = 1024


def parse_args() -> Dict:
  parser = argparse.ArgumentParser()
  parser.add_argument('--phenobench_dir', required=True, type=Path, help='Path to ground truth directory.')
  parser.add_argument('--prediction_dir', required=True, type=Path, help='Path to prediction directory.')
  parser.add_argument('--export', required=True, type=Path, help='Path to export directory.')
  parser.add_argument('--split',
                      choices=['train', 'val', 'test'],
                      required=True,
                      type=str,
                      help='Specify which split to use for evaluation.')

  parser.add_argument('--mapping', required=False, type=Path, help='Path to mapping.json file (only for test-split)')
  # Required directory structure
  # ├── ground-truth
  # │   ├── plant_instances
  # │   ├── plant_visibility
  # │   └── semantics
  # └── prediction
  #     └── plant_bboxes

  args = vars(parser.parse_args())

  args['export'].mkdir(parents=True, exist_ok=True)

  if args['split'] == 'test':
    if (not args['mapping']):
      raise TypeError(
          f"For the evaluation on the test-split you need to specify the '--date' and '--mapping' arguments.")

  return args


def evaluate_plant_detection(args) -> Dict[str, float]:

  # ------- Ground Truth -------
  gt_instance_fnames = get_png_files_in_dir(args['phenobench_dir'] / args['split'] / 'plant_instances')
  gt_semantic_fnames = get_png_files_in_dir(args['phenobench_dir'] / args['split'] / 'semantics')
  gt_visibility_fnames = get_png_files_in_dir(args['phenobench_dir'] / args['split'] / 'plant_visibility')

  # ------- Predictions -------
  pred_bboxes_fnames = get_txt_files_in_dir(args['prediction_dir'] / 'plant_bboxes')

  # ------- Load the mapping from randomized to original fnames -------
  if args['split'] == 'test':
    with open(args['mapping'], 'r') as istream:
      fname_mapping = json.load(istream)
    fname_mapping_reverse = {val: key for key, val in fname_mapping.items()}

    pred_bboxes_original_fnames = []
    for pred_fname in pred_bboxes_fnames:
      pred_bboxes_original_fnames.append(fname_mapping[pred_fname])

    gt_instance_fnames.sort()
    gt_semantic_fnames.sort()
    gt_visibility_fnames.sort()
    pred_bboxes_original_fnames.sort()
  else:
    pred_bboxes_original_fnames = pred_bboxes_fnames
    fname_mapping_reverse = {fname: fname for fname in pred_bboxes_fnames}

  # ------- Setup evaluator -------
  evaluator = MeanAveragePrecision(box_format='cxcywh', class_metrics=True)
  if args['split'] == 'test':
    evaluator_2020 = MeanAveragePrecision(box_format='cxcywh', class_metrics=True)
    evaluator_2021 = MeanAveragePrecision(box_format='cxcywh', class_metrics=True)

  for gt_instance_fname, gt_semantic_fname, gt_visibility_fname, pred_bboxes_fnames in tqdm(
      zip(gt_instance_fnames, gt_semantic_fnames, gt_visibility_fnames, pred_bboxes_original_fnames),
      total=len(gt_instance_fnames)):
    assert gt_instance_fname.split('.')[0] == gt_semantic_fname.split('.')[0] == gt_visibility_fname.split(
        '.')[0] == pred_bboxes_fnames.split('.')[0]

    gt_instance_map = load_file_as_tensor(args['phenobench_dir'] / args['split'] / 'plant_instances' /
                                          gt_instance_fname).squeeze()  # [H x W]
    gt_semantics = load_file_as_tensor(args['phenobench_dir'] / args['split'] / 'semantics' /
                                       gt_semantic_fname).squeeze()  # [H x W]
    gt_visibility = load_file_as_tensor(args['phenobench_dir'] / args['split'] / 'plant_visibility' /
                                        gt_instance_fname).squeeze()  # [H x W]
    targets = cvt_gt_to_bbox_map(gt_instance_map, gt_semantics, gt_visibility)

    yolo_pred = load_yolo_txt_file(args['prediction_dir'] / 'plant_bboxes' / fname_mapping_reverse[pred_bboxes_fnames])
    preds = cvt_yolo_to_bbox_map(yolo_pred)

    # Remove predictions and ground truths that belong to partially visible instances
    filter_partials_boxes(preds[0], targets[0])
    if args['split'] == 'test':
      if '2021' not in gt_instance_fname:
        evaluator_2020.update(preds, targets)
      if '2021' in gt_instance_fname:
        evaluator_2021.update(preds, targets)

    evaluator.update(preds, targets)

  metrics = evaluator.compute()
  eval_results = {}
  if args['split'] == 'test':
    ext = "_all"
  else:
    ext = ""
  eval_results[f'mAP{ext}'] = round(float(metrics['map'] * 100), 2)
  eval_results[f'mAP_50{ext}'] = round(float(metrics['map_50'] * 100), 2)
  eval_results[f'mAP_75{ext}'] = round(float(metrics['map_75'] * 100), 2)
  eval_results[f'mAP_cls{ext}'] = [round(float(val * 100), 2) for val in metrics['map_per_class']]

  if "export" in args:
    fpath_out_all = args['export'] / 'all'
    fpath_out_all.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_all / 'eval_plant_bboxes.yaml', 'w') as ostream:
      yaml.dump(eval_results, ostream)

  if args['split'] == 'test':
    metrics_2020 = evaluator_2020.compute()
    eval_results_2020 = {}
    ext = "_2020"
    eval_results_2020[f'mAP{ext}'] = round(float(metrics_2020['map'] * 100), 2)
    eval_results_2020[f'mAP_50{ext}'] = round(float(metrics_2020['map_50'] * 100), 2)
    eval_results_2020[f'mAP_75{ext}'] = round(float(metrics_2020['map_75'] * 100), 2)
    eval_results_2020[f'mAP_cls{ext}'] = [round(float(val * 100), 2) for val in metrics_2020['map_per_class']]

    if "export" in args:
      fpath_out_2020 = args['export'] / '2020'
      fpath_out_2020.mkdir(parents=True, exist_ok=True)
      with open(fpath_out_2020 / 'eval_plant_bboxes.yaml', 'w') as ostream:
        yaml.dump(eval_results_2020, ostream)

    metrics_2021 = evaluator_2021.compute()
    eval_results_2021 = {}
    ext = "_2021"
    eval_results_2021[f'mAP{ext}'] = round(float(metrics_2021['map'] * 100), 2)
    eval_results_2021[f'mAP_50{ext}'] = round(float(metrics_2021['map_50'] * 100), 2)
    eval_results_2021[f'mAP_75{ext}'] = round(float(metrics_2021['map_75'] * 100), 2)
    eval_results_2021[f'mAP_cls{ext}'] = [round(float(val * 100), 2) for val in metrics_2021['map_per_class']]

    fpath_out_2021 = args['export'] / '2021'
    fpath_out_2021.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_2021 / 'eval_plant_bboxes.yaml', 'w') as ostream:
      yaml.dump(eval_results_2021, ostream)

  return eval_results


if __name__ == '__main__':
  args = parse_args()
  evaluate_plant_detection(args)



#!/usr/bin/env python3

import argparse
from pathlib import Path

from phenobench.evaluation.auxiliary.common import centered_text
from phenobench.evaluation.evaluate_leaf_bounding_boxes import evaluate_leaf_detection
from phenobench.evaluation.evaluate_leaf_instance_masks_panoptic import (
    evaluate_leaf_instances,
)
from phenobench.evaluation.evaluate_plant_bounding_boxes import evaluate_plant_detection
from phenobench.evaluation.evaluate_plant_instance_masks_panoptic import (
    evaluate_plant_instances,
)
from phenobench.evaluation.evaluate_semantics import evaluate_semantics


def run():
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--task',
      choices=['semantics', 'panoptic', 'leaf_instances', 'plant_detection', 'leaf_detection', 'hierarchical'],
      required=True,
      type=str,
      help='task to evaluate.')
  parser.add_argument('--phenobench_dir', required=True, type=Path, help='Path to the root directory of PhenoBench.')
  parser.add_argument('--prediction_dir', required=True, type=Path, help='Path to prediction directory.')
  parser.add_argument('--split',
                      choices=['train', 'val'],
                      required=True,
                      type=str,
                      default='val',
                      help='Specify which split to use for evaluation.')
  args = vars(parser.parse_known_args()[0])

  if args["task"] == "semantics":
    print(centered_text("Computing metrics for semantic segmentation"))

    results = evaluate_semantics(args)

    print(centered_text(f"Metrics for {args['split']} split of PhenoBench"))

    print(f"{'IoU (crop)':10}: {results['crop']:4}")
    print(f"{'IoU (weed)':10}: {results['weed']:4}")
    print(f"{'mIoU':10}: {results['mIoU']:4}")

  elif args["task"] == "panoptic":
    print(centered_text("Computing metrics for plant panoptic segmentation"))

    semantic_results = evaluate_semantics(args)
    instance_results = evaluate_plant_instances(args)

    print(centered_text(f"Metrics for {args['split']} split of PhenoBench"))

    iou_soil = semantic_results['soil']
    pq_crop = instance_results['plants_cls'][1]['pq']
    pq_weed = instance_results['plants_cls'][2]['pq']

    print(f"{'IoU (soil)':10}: {iou_soil:4}")
    print(f"{'PQ (crop)':10}: {pq_crop:4}")
    print(f"{'PQ (weed)':10}: {pq_weed:4}")
    print(f"{'PQ+':10}: {(iou_soil+pq_crop+pq_weed)/3:4.2f}")

  elif args["task"] == "leaf_instances":
    print(centered_text("Computing metrics for leaf instance segmentation"))

    results = evaluate_leaf_instances(args)

    print(centered_text(f"Metrics for {args['split']} split of PhenoBench"))
    pq_leaf = results["leaves_pq"]
    print(f"{'PQ (leaf)':10}: {pq_leaf:4}")

  elif args["task"] == "plant_detection":
    print(centered_text("Computing metrics for plant detection"))

    results = evaluate_plant_detection(args)

    print(centered_text(f"Metrics for {args['split']} split of PhenoBench"))

    print(f"{'mAP':10}: {results['mAP']:4}")
    print(f"{'mAP_50':10}: {results['mAP_50']:4}")
    print(f"{'mAP_75':10}: {results['mAP_75']:4}")
    print(f"{'AP (crop)':10}: {results['mAP_cls'][0]:4}")
    print(f"{'AP (weed)':10}: {results['mAP_cls'][1]:4}")

  elif args["task"] == "leaf_detection":
    print(centered_text("Computing metrics for leaf detection"))

    results = evaluate_leaf_detection(args)

    print(centered_text(f"Metrics for {args['split']} split of PhenoBench"))

    print(f"{'mAP':10}: {results['mAP']:4}")
    print(f"{'mAP_50':10}: {results['mAP_50']:4}")
    print(f"{'mAP_75':10}: {results['mAP_75']:4}")

  elif args["task"] == "hierarchical":
    print(centered_text("Computing metrics for leaf detection"))

    semantic_results = evaluate_semantics(args)
    instance_results = evaluate_plant_instances(args)
    leaf_results = evaluate_leaf_instances(args)

    print(centered_text(f"Metrics for {args['split']} split of PhenoBench"))

    iou_soil = semantic_results['soil']
    iou_weed = semantic_results['weed']

    pq_leaf = leaf_results["leaves_pq"]
    pq_crop = instance_results['plants_cls'][1]['pq']

    print(f"{'IoU (soil)':10}: {iou_soil:4}")
    print(f"{'IoU (weed)':10}: {iou_weed:4}")
    print(f"{'PQ (leaf)':10}: {pq_leaf:4}")
    print(f"{'PQ (crop)':10}: {pq_crop:4}")
    print(f"{'PQ':10}: {(pq_crop+pq_leaf)/2:4.2f}")
    print(f"{'PQ+':10}: {(iou_soil+iou_weed+pq_crop+pq_leaf)/4:4.2f}")


if __name__ == '__main__':
  run()


#!/usr/bin/env python3

import argparse
import os
from zipfile import ZipFile

import numpy as np


class ValidationException(Exception):
  pass


def run():
  parser = argparse.ArgumentParser(
      description=
      "Validate a submission zip file needed to evaluate on CodaLab competitions of PhenoBench.\n\nThe validator checks:\n  1. correct folder structure,\n  2. existence of label files for each image.\nInvalid labels are ignored by the evaluation script, therefore we don't check\nfor invalid labels.",
      formatter_class=argparse.RawTextHelpFormatter)

  parser.add_argument(
      "--task",
      type=str,
      required=True,
      choices=["semantics", "panoptic", "leaf_instances", "plant_detection", "leaf_detection", "hierarchical"],
      help='task for which the zip file should be validated.')

  parser.add_argument(
      "--zipfile",
      type=str,
      required=True,
      help='zip file that should be validated.',
  )

  parser.add_argument('--phenobench_dir',
                      type=str,
                      required=True,
                      help='Root directory containing the folder train, val, test of the PhenoBench.')

  FLAGS, _ = parser.parse_known_args()

  checkmark = "\u2713"

  required_folders = {
      "semantics": ["semantics"],
      "panoptic": ["semantics", "plant_instances"],
      "leaf_instances": ["leaf_instances"],
      "plant_detection": ["plant_bboxes"],
      "leaf_detection": ["leaf_bboxes"],
      "hierarchical": ["semantics", "plant_instances", "leaf_instances"]
  }

  try:

    print('Validating zip archive "{}".\n'.format(FLAGS.zipfile))

    print(" ========== {:^15} ========== ".format(FLAGS.task))

    print("  1. Checking filename.............. ", end="", flush=True)
    if not FLAGS.zipfile.endswith('.zip'):
      raise ValidationException('Competition submission must end with ".zip"')
    print(checkmark)

    with ZipFile(FLAGS.zipfile) as zipfile:

      print("  2. Checking directory structure... ", end="", flush=True)

      directories = [folder.filename for folder in zipfile.infolist() if folder.filename.endswith("/")]
      for expected_folder in required_folders[FLAGS.task]:
        if expected_folder + "/" not in directories:
          raise ValidationException(
              f'Directory "{expected_folder}" missing inside zip file. Your zip file should contain following folders: {", ".join(required_folders[FLAGS.task])}'
          )

      print(checkmark)

      print('  3. Checking files................. ', end='', flush=True)

      prediction_files = {info.filename: info for info in zipfile.infolist() if not info.filename.endswith("/")}
      image_files = sorted(os.listdir(os.path.join(FLAGS.phenobench_dir, "test", "images")))

      for expected_folder in required_folders[FLAGS.task]:
        if "bboxes" in expected_folder:
          for image_file in image_files:
            expected_file = f"{expected_folder}/{os.path.splitext(image_file)[0]}.txt"
            if expected_file not in prediction_files:
              raise ValidationException(f'Missing prediction for {image_file} in folder {expected_folder}!')
        else:
          for image_file in image_files:
            expected_file = f"{expected_folder}/{image_file}"
            if expected_file not in prediction_files:
              raise ValidationException(f'Missing prediction for {image_file} in folder {expected_folder}!')

      print(checkmark)

  except ValidationException as ex:
    print("\n\n  " + "\u001b[1;31m>>> Error: " + str(ex) + "\u001b[0m")
    exit(1)

  print("\n\u001b[1;32mEverything ready for submission!\u001b[0m  \U0001F389")


if __name__ == "__main__":
  run()



from typing import Dict, List, Tuple

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from PIL import Image


def draw_semantics(axes: Axes,
                   image: Image,
                   semantics: np.array,
                   alpha: float = 0.5,
                   colors: Dict[int, Tuple[int, int, int]] = {
                       1: (0, 255, 0),
                       2: (255, 0, 0),
                       3: (0, 255, 255),
                       4: (255, 0, 255)
                   },
                   mask_classes: List[int] = [0]):
  """ draw pixel-wise semantics according to given image and semantics. 

  Args:
      axes (matplotlib.axes.Axes): axes object to draw into.
      image (PIL.Image): image that should be drawn.
      semantics (np.array): semantics (1=crop, 2=weed, 3=partial crop, 4=partial weed)
      alpha(float): transparency of semantic mask drawn on the image.
      colors(Dict[int, tuple]): mapping of class ids to colors. Default: 1(crop) = (0,255,0), 2(weed) = (0,255,0), 3(partial_crop)= (0,255,255), 4(partial_weed) = (255,0,255)
      mask_classes (List[int]): class indices that are not shown in the final composition.
  """

  im_semantics = np.zeros((*semantics.shape, 3), dtype=float)

  for label, color in colors.items():
    im_semantics[semantics == label] = np.array(color) / 255.0

  im_alpha = np.ones((*semantics.shape, 3), dtype=float) * alpha
  for c in mask_classes:
    im_alpha[semantics == c] = 0.0

  composed_image = (1.0 - im_alpha) * (np.array(image) / 255.0) + im_alpha * im_semantics
  axes.imshow(composed_image)

  # not sure why, but `ax.im_show(im_semantics, alpha=im_alpha)` doesn't work as expected:
  # im_alpha = np.ones(semantics.shape, dtype=float) * alpha
  # for c in mask_classes: im_alpha[semantics == c] = 0.0
  # axes.imshow(image)
  # axes.imshow(im_semantics, alpha = im_alpha)


def draw_instances(ax: Axes, image: Image, instance_mask: np.array, alpha: float = 0.5):
  """ Draw instance masks from provided data, where individual masks are indicated by random colors.

  Args:
      ax (Axes): axes object to draw into.
      image (Image): image that should be drawn.
      instance_masks (np.array): _description_
      alpha (float, optional): _description_. Defaults to 0.5.
  """
  random_colors = plt.get_cmap('tab10').resampled(10)
  unique_ids = np.unique(instance_mask)
  # unique_ids = unique_ids[unique_ids > 0]

  im_mask = np.zeros((*instance_mask.shape, 3))
  for idx, uid in enumerate(unique_ids):
    im_mask[instance_mask == uid] = random_colors(idx % 10)[:3]

  im_alpha = np.ones((*instance_mask.shape, 3)) * alpha
  im_alpha[instance_mask == 0, :] = 0.0

  float_image = (np.array(image) / 255.0)
  composed_image = (1.0 - im_alpha) * float_image + im_alpha * im_mask

  ax.imshow(composed_image)


def draw_bboxes(ax: Axes,
                image: Image,
                bboxes: List[Dict],
                colors: Dict[int, Tuple[int, int, int]] = {
                    1: (0, 255, 0),
                    2: (255, 0, 0),
                    3: (0, 255, 255),
                    4: (255, 0, 255)
                },
                linewidth=2):
  """ Draw the given bounding boxes on the image.

  Args:
      ax (Axes): axes object to draw into.
      image (Image): image that should be drawn.
      bboxes (List[Dict): list of bounding boxes in the format {"label":<label>, "center": (x,y), "width": w, "height": h},
      colors(Dict[int, tuple]): mapping of class ids to colors. Default: 1(crop) = (0,255,0), 2(weed) = (0,255,0), 3(partial_crop)= (0,255,255), 4(partial_weed) = (255,0,255)
      linewidth(int): thickness of the lines of the bounding box. Default: 2
  """

  ax.imshow(image)

  for bbox in bboxes:
    x, y = bbox["center"][0] - bbox["width"] / 2, bbox["center"][1] - bbox["height"] / 2
    rect = patches.Rectangle((x, y),
                             bbox["width"],
                             bbox["height"],
                             linewidth=linewidth,
                             edgecolor=np.array(colors[bbox["label"]]) / 255.0,
                             facecolor='none')
    ax.add_patch(rect)



from typing import Dict

import torch

from .convert import bounding_box_from_cxcywh_to_xyxy, box_to_mask


def filter_partial_masks(pred_instance_masks: torch.Tensor, pred_semantics: torch.Tensor, gt_instance_masks: torch.Tensor, gt_semantics: torch.Tensor, gt_visibility: torch.Tensor):
  assert torch.max(gt_visibility) <= 1.0
  assert torch.min(gt_visibility) >= 0.0

  gt_instance_partial_masks = []
  # set semantics of partial masks to 0 and store partial_masks
  for gt_instance_id in torch.unique(gt_instance_masks):
    if gt_instance_id == 0:
      continue
    gt_mask = gt_instance_masks == gt_instance_id
    gt_vis = torch.unique(gt_visibility[gt_mask]) # contain a single value
    if gt_vis > 0.5:
      continue
    gt_semantics[gt_mask] = 0
    gt_instance_partial_masks.append(gt_mask)


  for pred_instance_id in torch.unique(pred_instance_masks):
    if pred_instance_id == 0:
      continue
    pred_mask = pred_instance_masks == pred_instance_id

    for gt_mask in gt_instance_partial_masks:
      # compute how much of the prediction is within the ground truth
      a = torch.sum(gt_mask & pred_mask)
      b = torch.sum(pred_mask)
      score = a / (b + 1e-12)
      assert score <= 1.0

      if score > 0.5:
        pred_semantics[pred_mask] = 0

def filter_partials_boxes(preds: Dict, targets: Dict):
  n_preds = preds['boxes'].shape[0]
  n_targets = targets['boxes'].shape[0]

  remove_pred_indicies = []
  remove_gt_indicies = []

  for i in range(n_preds):
    bbox_pred = preds['boxes'][i, :]
    cx, cy, width, height = bbox_pred[0].item(), bbox_pred[1].item(), bbox_pred[2].item(), bbox_pred[3].item()
    xmin, xmax, ymin, ymax = bounding_box_from_cxcywh_to_xyxy(cx, cy, width, height)
    bbox_pred_canvas = box_to_mask(xmin, xmax, ymin, ymax)

    for j in range(n_targets):
      vis = targets['visibility'][j]
      if vis > 0.5:
        continue
      remove_gt_indicies.append(j)

      bbox_target = targets['boxes'][j, :]
      cx, cy, width, height = bbox_target[0].item(), bbox_target[1].item(), bbox_target[2].item(), bbox_target[3].item()
      xmin, xmax, ymin, ymax = bounding_box_from_cxcywh_to_xyxy(cx, cy, width, height)
      bbox_gt_canvas = box_to_mask(xmin, xmax, ymin, ymax)

      # compute how much of the prediction is within the ground truth
      a = torch.sum(bbox_pred_canvas & bbox_gt_canvas)
      b = torch.sum(bbox_pred_canvas)
      score = a / (b + 1e-12)
      assert score <= 1.0

      if score > 0.5:
        remove_pred_indicies.append(i)

  # filter predictions
  mask_pred = torch.ones_like(preds['labels'], dtype=torch.bool)
  for rm_idx in remove_pred_indicies:
    mask_pred[rm_idx] = False
  preds['labels'] = preds['labels'][mask_pred]
  preds['scores'] = preds['scores'][mask_pred]
  preds['boxes'] = preds['boxes'][mask_pred]

  # filter targets
  mask_target = torch.ones_like(targets['labels'], dtype=torch.bool)
  for rm_idx in remove_gt_indicies:
    mask_target[rm_idx] = False
  targets['labels'] = targets['labels'][mask_target]
  targets['boxes'] = targets['boxes'][mask_target]
  targets['visibility'] = targets['visibility'][mask_target]



from typing import Dict, List, Tuple

import torch

IMG_WIDTH = 1024
IMG_HEIGHT = 1024

def convert_partial_semantics(semantics: torch.Tensor) -> torch.Tensor:
  semantics_converted = semantics.clone()

  # convert partial crops to regular crops
  mask_partial_crops = semantics == 3
  semantics_converted[mask_partial_crops] = 1

  # convert partial weeds to regular weeds
  mask_partial_weeds = semantics == 4
  semantics_converted[mask_partial_weeds] = 2

  return semantics_converted

def bounding_box_from_mask_xyxy(mask: torch.Tensor) -> Tuple[int, int, int, int]:
  rows, cols = mask.nonzero(as_tuple=True)

  x_min = int(torch.min(cols))
  y_min = int(torch.min(rows))
  x_max = int(torch.max(cols))
  y_max = int(torch.max(rows))

  return x_min, y_min, x_max, y_max

def bounding_box_from_mask_cxcywh(mask: torch.Tensor) -> Tuple[int, int, int, int]:

  x_min, y_min, x_max, y_max = bounding_box_from_mask_xyxy(mask)

  cx = int(round((x_min + x_max) / 2))
  cy = int(round((y_min + y_max) / 2))
  width = x_max - x_min
  height = y_max - y_min

  assert width >= 0
  assert height >= 0

  return cx, cy, width, height

def bounding_box_from_cxcywh_to_xyxy(cx: int, cy: int, width: int, height: int) -> Tuple[int, int, int, int]:
  x_min = round(cx - (width / 2))
  x_max = round(cx + (width / 2))
  y_min = round(cy - (height / 2))
  y_max = round(cy + (height / 2))

  return x_min, x_max, y_min, y_max

def box_to_mask(x_min:int, x_max: int, y_min: int, y_max: int) -> torch.Tensor:
  canvas = torch.zeros((IMG_HEIGHT, IMG_WIDTH), dtype=torch.bool)

  canvas[y_min: (y_max + 1), x_min: (x_max + 1)] = True

  return canvas

def cvt_yolo_to_bbox_map(yolo_anno: torch.Tensor) -> List[Dict[str, torch.Tensor]]:
  """ Convert YOLO annotations to format required by MAP of torchmetrics.

  Please check the following to get more information: https://torchmetrics.readthedocs.io/en/stable/detection/mean_average_precision.html

  Args:
      yolo_anno (torch.Tensor): annotations provided by YOLO ... of shape [n_preds x 6]
        - 1st col -> label id
        - 2nd col -> center_x in [0,1]
        - 3rd col -> center_y in [0,1]
        - 4th col -> width in [0,1]
        - 5th col -> height in [0,1]
        - 6th col -> confidence score in [0,1]

  Returns:
      List[Dict[str, torch.Tensor]]: format required by MAP of torchmetrics.
  """

  if len(yolo_anno.shape) == 1:
    yolo_anno = yolo_anno.unsqueeze(dim=0)

  n_predicitions = yolo_anno.shape[0]
  boxes = []
  labels = []
  scores = []
  for i in range(n_predicitions):
    # ------- Label -------
    try:
      label = int(yolo_anno[i, 0]) # + 1
    except IndexError:
      continue
    if label == 0:
      continue

    # ------- Bounding Box -------
    cx = int(round(float(yolo_anno[i, 1]) * IMG_WIDTH))
    cy = int(round(float(yolo_anno[i, 2]) * IMG_HEIGHT))
    width = int(round(float(yolo_anno[i, 3]) * IMG_WIDTH))
    height = int(round(float(yolo_anno[i, 4]) * IMG_HEIGHT))

    # ------- Confidence Score -------
    score = float(yolo_anno[i, 5])

    # ------- Accumulate -------
    labels.append(torch.Tensor([label]))
    boxes.append(torch.Tensor([cx, cy, width, height]))
    scores.append(torch.Tensor([score]))

  out = {}
  try:
    out['labels'] = torch.stack(labels).squeeze(1).type(torch.uint8)  # [n_objects]
    out['boxes'] = torch.stack(boxes)  # [n_objects x 4]
    out['scores'] = torch.stack(scores).squeeze(1)  # [n_objects]
  except RuntimeError:
    # i.e., there isn't any prediction
    out['labels'] = torch.tensor([])
    out['boxes'] = torch.tensor([])
    out['scores'] = torch.tensor([])


  return [out]


def cvt_gt_to_bbox_map(instance_map: torch.Tensor, semantics: torch.Tensor, visibility: torch.Tensor) -> List[Dict[str, torch.Tensor]]:
  assert torch.max(visibility) <= 1.0
  assert torch.min(visibility) >= 0.0

  instance_ids = torch.unique(instance_map)
  instance_ids = instance_ids[instance_ids != 0]  # ignore background

  boxes = []
  labels = []
  visibilities = []
  for instance_id in instance_ids:
    # ------- Mask -------
    instance_mask = instance_map == instance_id

    # ------- Bounding Box -------
    cx, cy, width, height = bounding_box_from_mask_cxcywh(instance_mask)

    # ------- Label -------
    label = torch.unique(semantics[instance_mask])
    try:
      if label == 0:
        continue
      if len(label) != 1:
        assert False, "Single instance has multiple semantic classes"
    except RuntimeError:
      continue

    # ------- Visibility -------
    vis = torch.unique(visibility[instance_mask])

    # ------- Accumulate -------
    boxes.append(torch.Tensor([cx, cy, width, height]))
    labels.append(torch.Tensor([label]))
    visibilities.append(torch.Tensor(vis))

  out = {}
  out['labels'] = torch.stack(labels).squeeze(1).type(torch.uint8)  # [n_objects]
  out['boxes'] = torch.stack(boxes)  # [n_objects x 4]
  out['visibility'] = torch.stack(visibilities)  # [n_objects]

  return [out]



import os
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import torch
import torchvision.transforms as T
from numpy import genfromtxt
from PIL import Image


def is_png(filename: str) -> bool:
  return filename.endswith('.png')


def is_txt(filename: str) -> bool:
  return filename.endswith('.txt')


def get_png_files_in_dir(path_to_dir: Path) -> List[str]:
  filenames = sorted([filename for filename in os.listdir(path_to_dir) if is_png(filename)])

  return filenames


def get_txt_files_in_dir(path_to_dir: Path) -> List[str]:
  filenames = sorted([filename for filename in os.listdir(path_to_dir) if is_txt(filename)])

  return filenames

def load_file_as_int_tensor(path_to_file: Path) -> torch.Tensor:

  tensor = Image.open(path_to_file)
  tensor = np.asarray(tensor).astype(np.int32)
  tensor = torch.Tensor(tensor).type(torch.int32)

  return tensor

def load_file_as_tensor(path_to_file: Path) -> torch.Tensor:
  to_tensor = T.ToTensor()

  tensor = Image.open(path_to_file)
  tensor = to_tensor(tensor)

  return tensor


def load_yolo_txt_file(path_to_file: Path) -> torch.Tensor:
  annos = genfromtxt(path_to_file)

  return torch.Tensor(annos)

def centered_text(text: str, line_width: int = 80, fill: str = '=') -> str:
  """ generate string such that text is center in given line width using the fill """
  fill_width = (line_width - len(text)-2)//2
  repeat = fill_width // len(fill)

  return f"{fill*repeat} {text} {fill*repeat}"



import argparse
import json
from pathlib import Path
from typing import Dict

import yaml
from tqdm import tqdm

from phenobench.evaluation.auxiliary.common import (
    get_png_files_in_dir,
    load_file_as_int_tensor,
    load_file_as_tensor,
)
from phenobench.evaluation.auxiliary.filter import filter_partial_masks
from phenobench.evaluation.auxiliary.panoptic_eval import PanopticEvaluator


def parse_args() -> Dict:
  parser = argparse.ArgumentParser()
  parser.add_argument('--phenobench_dir', required=True, type=Path, help='Path to ground truth directory.')
  parser.add_argument('--prediction_dir', required=True, type=Path, help='Path to prediction directory.')
  parser.add_argument('--export', required=False, type=Path, help='Path to export directory.')
  parser.add_argument('--split',
                      choices=['train', 'val', 'test'],
                      required=True,
                      type=str,
                      help='Specify which split to use for evaluation.')
  parser.add_argument('--mapping', required=False, type=Path, help='Path to mapping.json file (only for test-split)')

  # Required directory structure
  # ├── ground-truth
  # |   └── <split>
  # │       ├── plant_instances
  # |       ├── plant_visibility
  # │       └── semantics
  # └── prediction
  #     └── plant_instances
  #     └── semantics

  args = vars(parser.parse_args())
  args['export'].mkdir(parents=True, exist_ok=True)

  if args['split'] == 'test':
    if (not args['mapping']):
      raise TypeError(
          f"For the evaluation on the test-split you need to specify the '--date' and '--mapping' arguments.")

  return args


def evaluate_plant_instances(args: Dict[str, str]) -> Dict[str, float]:

  # ------- Ground Truth -------
  gt_instance_fnames = get_png_files_in_dir(args['phenobench_dir'] / args['split'] / 'plant_instances')
  gt_semantic_fnames = get_png_files_in_dir(args['phenobench_dir'] / args['split'] / 'semantics')
  gt_visibility_fnames = get_png_files_in_dir(args['phenobench_dir'] / args['split'] / 'plant_visibility')

  # ------- Predictions -------
  pred_instance_fnames = get_png_files_in_dir(args['prediction_dir'] / 'plant_instances')
  pred_semantic_fnames = get_png_files_in_dir(args['prediction_dir'] / 'semantics')

  # ------- Load the mapping from randomized to original fnames -------
  if args['split'] == 'test':
    with open(args['mapping'], 'r') as istream:
      fname_mapping = json.load(istream)
    fname_mapping_reverse = {val: key for key, val in fname_mapping.items()}

    pred_instance_original_fnames = []
    pred_semantic_original_fnames = []
    for pred_inst_fname, pred_sem_fname in zip(pred_instance_fnames, pred_semantic_fnames):
      pred_instance_original_fnames.append(fname_mapping[pred_inst_fname])
      pred_semantic_original_fnames.append(fname_mapping[pred_sem_fname])
    gt_instance_fnames.sort()
    gt_semantic_fnames.sort()
    gt_visibility_fnames.sort()
    pred_instance_original_fnames.sort()
    pred_semantic_original_fnames.sort()
  else:
    pred_instance_original_fnames = pred_instance_fnames
    pred_semantic_original_fnames = pred_semantic_fnames
    fname_mapping_reverse = {fname: fname for fname in pred_instance_fnames}

  # ------- Setup evaluator -------
  evaluator = PanopticEvaluator()
  if args['split'] == 'test':
    evaluator_2020 = PanopticEvaluator()
    evaluator_2021 = PanopticEvaluator()

  n_total = len(gt_instance_fnames)
  for gt_instance_fname, gt_semantic_fname, gt_visibility_fname, pred_instance_fname, pred_semantic_fname, in tqdm(
      zip(gt_instance_fnames, gt_semantic_fnames, gt_visibility_fnames, pred_instance_original_fnames,
          pred_semantic_original_fnames),
      total=n_total):
    assert gt_instance_fname == gt_semantic_fname == gt_visibility_fname == pred_instance_fname == pred_semantic_fname

    gt_instance_map = load_file_as_tensor(args['phenobench_dir'] / args['split'] / 'plant_instances' /
                                          gt_instance_fname).squeeze()  # [H x W]
    gt_semantics = load_file_as_tensor(args['phenobench_dir'] / args['split'] / 'semantics' /
                                       gt_semantic_fname).squeeze()  # [H x W]
    gt_visibility = load_file_as_tensor(args['phenobench_dir'] / args['split'] / 'plant_visibility' /
                                        gt_instance_fname).squeeze()  # [H x W]

    pred_instance_map = load_file_as_int_tensor(args['prediction_dir'] / 'plant_instances' /
                                                fname_mapping_reverse[pred_instance_fname]).squeeze()  # [H x W]
    pred_semantics = load_file_as_int_tensor(args['prediction_dir'] / 'semantics' /
                                             fname_mapping_reverse[pred_semantic_fname]).squeeze()  # [H x W]

    filter_partial_masks(pred_instance_map, pred_semantics, gt_instance_map, gt_semantics, gt_visibility)
    if args['split'] == 'test':
      if '2021' not in gt_instance_fname:
        evaluator_2020.update(pred_semantics, gt_semantics, pred_instance_map, gt_instance_map)
      if '2021' in gt_instance_fname:
        evaluator_2021.update(pred_semantics, gt_semantics, pred_instance_map, gt_instance_map)

    evaluator.update(pred_semantics, gt_semantics, pred_instance_map, gt_instance_map)

  metrics = round(float(evaluator.compute() * 100), 2)
  metrics_classwise = evaluator.results_classwise
  for class_id in metrics_classwise:
    metrics_classwise[class_id]['pq'] = round(float(metrics_classwise[class_id]['pq'] * 100), 2)

  eval_results = {}
  if args['split'] == 'test':
    ext = "_all"
  else:
    ext = ""

  eval_results[f"plants_pq{ext}"] = metrics
  eval_results[f"plants_cls{ext}"] = metrics_classwise

  if "export" in args:
    fpath_out_all = args['export'] / 'all'
    fpath_out_all.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_all / 'eval_plant_pq.yaml', 'w') as ostream:
      yaml.dump(eval_results, ostream)

  if args['split'] == 'test':
    # 2020
    metrics_2020 = round(float(evaluator_2020.compute() * 100), 2)
    metrics_2020_classwise = evaluator_2020.results_classwise
    for class_id in metrics_2020_classwise:
      metrics_2020_classwise[class_id]['pq'] = round(float(metrics_2020_classwise[class_id]['pq'] * 100), 2)

    eval_results_2020 = {}
    ext = "_2020"
    eval_results_2020[f"plants_pq{ext}"] = metrics_2020
    eval_results_2020[f"plants_cls{ext}"] = metrics_2020_classwise

    fpath_out_2020 = args['export'] / '2020'
    fpath_out_2020.mkdir(parents=True, exist_ok=True)
    with open(fpath_out_2020 / 'eval_plant_pq.yaml', 'w') as ostream:
      yaml.dump(eval_results_2020, ostream)

    # 2021
    metrics_2021 = round(float(evaluator_2021.compute() * 100), 2)
    metrics_2021_classwise = evaluator_2021.results_classwise
    for class_id in metrics_2021_classwise:
      metrics_2021_classwise[class_id]['pq'] = round(float(metrics_2021_classwise[class_id]['pq'] * 100), 2)

    eval_results_2021 = {}
    ext = "_2021"
    eval_results_2021[f"plants_pq{ext}"] = metrics_2021
    eval_results_2021[f"plants_cls{ext}"] = metrics_2021_classwise
    if "export" in args:
      fpath_out_2021 = args['export'] / '2021'
      fpath_out_2021.mkdir(parents=True, exist_ok=True)
      with open(fpath_out_2021 / 'eval_plant_pq.yaml', 'w') as ostream:
        yaml.dump(eval_results_2021, ostream)

  return eval_results

if __name__ == '__main__':
  args = parse_args()
  evaluate_plant_instances(args)



import numpy as np
import torch


class PanopticEvaluator():
  def __init__(self):
    self.evaluator = PanopticQuality()

  def update(self, semantic_pred, semantic_gt, instance_pred, instance_gt):
    self.evaluator.compute_pq(semantic_pred, semantic_gt, instance_pred, instance_gt)

  def compute(self):
    pq = self.evaluator.average_pq(self.evaluator.panoptic_qualities)

    return pq

  @property
  def results_classwise(self):
    return self.evaluator.panoptic_qualities


class PanopticQuality:
  def __init__(self):
    self.panoptic_qualities = {}

  def reset(self):
    self.panoptic_qualities.clear()

  def compute_pq_single_class(self, prediction, groundtruth):
    numerator_iou = torch.tensor(0.)
    class_matches = 0
    unmatched_class_predictions = 0

    # take the non-zero INSTANCE labels for both prediction and groundtruth
    labels = torch.unique(prediction)
    gt_labels = torch.unique(groundtruth)
    labels = labels[labels > 0]
    gt_labels = gt_labels[gt_labels > 0]

    gt_masks = [groundtruth == gt_label for gt_label in gt_labels]
    gt_areas = [(groundtruth == gt_label).sum() for gt_label in gt_labels]

    not_matched = [True] * len(gt_labels)

    for label in labels: # for each predicted label
      iou = torch.tensor(0.)
      best_idx = 0

      pred_mask = (prediction == label)
      pred_area = (prediction == label).sum()

      for idx in np.where(not_matched)[0]:
        gt_mask = gt_masks[idx]
        gt_area = gt_areas[idx]
        # compute iou with all instance gt labels and store the best
        intersection = ((pred_mask & gt_mask).sum()).float()
        union = (pred_area + gt_area - intersection).float()

        iou_tmp = intersection / union
        if iou_tmp > iou:
          iou = iou_tmp
          best_idx = idx

      # if the best iou is above 0.5, store the match pred_label-gt_label-iou
      if iou > 0.5:
        class_matches += 1
        numerator_iou += iou
        not_matched[best_idx] = False
      else:
        # unmatched_class_predictions.append(label.item())
        unmatched_class_predictions += 1

    true_positives = class_matches # len(class_matches)     # TP = number of matches
    # FP = number of unmatched predictions
    false_positives = unmatched_class_predictions # len(unmatched_class_predictions)
    # FN = number of unmatched gt labels
    false_negatives = len(gt_labels) - class_matches #len(class_matches)
    panoptic_quality_one_class = numerator_iou / \
        (true_positives + 0.5 * false_positives + 0.5 * false_negatives)

    return panoptic_quality_one_class, class_matches

  def compute_pq(self, semantic_pred, semantic_gt, instance_pred, instance_gt):
    # take the semantic labels (except 0 = void label)
    semantic_gt_labels = torch.unique(semantic_gt).int()
    semantic_gt_labels = semantic_gt_labels[semantic_gt_labels > 0].int()

    if semantic_gt_labels.sum() == 0:
      if instance_pred.sum() == 0:
        return 1.0, {}
      else:
        return 0.0 , {}


    for gt_label in semantic_gt_labels:
      # take pred and gt where they have sem_class = label
      semantic_tmp = (semantic_pred == gt_label).int()
      semantic_gt_tmp = (semantic_gt == gt_label).int()

      # take pred and gt instances corresponding to the considered semantic class
      instance_tmp = (instance_pred * semantic_tmp).int()
      instance_gt_tmp = (instance_gt * semantic_gt_tmp).int()

      pq, _ = self.compute_pq_single_class(instance_tmp, instance_gt_tmp)

      # update: count = number of images that contains the class, pq is the overall panoptic quality of the images
      if gt_label.item() not in self.panoptic_qualities.keys():
        self.panoptic_qualities[gt_label.item()] = {}
        self.panoptic_qualities[gt_label.item()]['count'] = 1
        self.panoptic_qualities[gt_label.item()]['pq'] = pq.item()
      else:
        self.panoptic_qualities[gt_label.item()]['count'] += 1
        self.panoptic_qualities[gt_label.item()]['pq'] = ((self.panoptic_qualities[gt_label.item()]['pq'] * (
            self.panoptic_qualities[gt_label.item()]['count']-1) + pq) / self.panoptic_qualities[gt_label.item()]['count']).item()

    # panoptic_quality = self.average_pq(self.panoptic_qualities)
    # return panoptic_quality, self.panoptic_qualities

    return self.panoptic_qualities

  def average_pq(self, panoptic_qualities):
    # overall panoptic quality averaged from all classes
    pq = 0.
    n = 0
    for item in panoptic_qualities.keys():
      pq += panoptic_qualities[item]['pq']
      n += 1

    if n == 0:
      return -1
    pq /= n
    return pq

  def panoptic_quality_forward(self, batch_semantic_pred, batch_semantic_gt, batch_instance_pred, batch_instance_gt):
    batch_size = batch_semantic_pred.shape[0]
    for item in range(batch_size):
      semantic_gt = batch_semantic_gt[item]
      semantic_pred = batch_semantic_pred[item]
      instance_gt = batch_instance_gt[item]
      instance_pred = batch_instance_pred[item]
      # for each batch item
      pq, pqs = self.compute_pq(
          semantic_pred, semantic_gt, instance_pred, instance_gt)
    return pq, pqs


# class PanopticEvaluator():

#   def __init__(self, mode: PanopticEvalMode):
#     self.mode = mode

#     self.pq = PanopticQuality()
#     self.pq_list = []
#     self.class_pq_list = []
#     self.class_sq_list = []
#     self.class_rq_list = []

#   def update_pq(self, semantic_pred: torch.Tensor, semantic_gt: torch.Tensor, instance_pred: torch.Tensor, instance_gt: torch.Tensor):
#     self.pq.reset()
#     pq_class, sq_class, rq_class = self.pq.compute_pq(
#         semantic_pred,
#         semantic_gt,
#         instance_pred,
#         instance_gt,
#     )
#     # self.pq_list.append(pq_image)
#     self.class_pq_list.append(pq_class)
#     self.class_sq_list.append(sq_class)
#     self.class_rq_list.append(rq_class)

#   def update(self, semantic_pred: torch.Tensor, semantic_gt: torch.Tensor, instance_pred: torch.Tensor, instance_gt: torch.Tensor):
#     self.update_pq(semantic_pred, semantic_gt, instance_pred, instance_gt)

#   def compute(self):
#     metrics = {}
#     #pdb.set_trace()
#     class_pq = torch.tensor(self.class_pq_list).sum(dim=0) / len(self.class_pq_list)
#     class_sq = torch.tensor(self.class_sq_list).sum(dim=0) / len(self.class_sq_list)
#     class_rq = torch.tensor(self.class_rq_list).sum(dim=0) / len(self.class_rq_list)
#     self.class_pq_list = []
#     self.class_sq_list = []
#     self.class_rq_list = []
#     metrics["pq"] = class_pq.mean()
#     if self.mode == PanopticEvalMode.CROPS_WEEDS:
#       metrics["pq_crop"] = class_pq[0]
#       metrics["pq_weed"] = class_pq[1]
#       metrics["sq_crop"] = class_sq[0]
#       metrics["sq_weed"] = class_sq[1]
#       metrics["rq_crop"] = class_rq[0]
#       metrics["rq_weed"] = class_rq[1]
#     elif self.mode == PanopticEvalMode.LEAVES:
#       metrics["pq_leaves"] = class_pq[0]
#       metrics["sq_leaves"] = class_sq[0]
#       metrics["rq_leaves"] = class_rq[0]
#     elif self.mode == PanopticEvalMode.CROPS:
#       metrics["pq_crops"] = class_pq[0]
#       metrics["sq_crops"] = class_sq[0]
#       metrics["rq_crops"] = class_rq[0]
#     return metrics


# class PanopticQuality:

#   def __init__(self):
#     # initialization of a list of dictionaries with len = num classes
#     self.panoptic_qualities = {}

#   def reset(self):
#     self.panoptic_qualities.clear()

#   def compute_pq_single_class(self, prediction, groundtruth):
#     if prediction.sum() == 0 and groundtruth.sum() == 0:
#       # if no predictions and no plants in groundtruth return pq 1
#       return 1.0, 0
#     numerator_iou = torch.tensor(0.0)
#     class_matches = []
#     unmatched_class_predictions = []

#     # take the non-zero INSTANCE labels for both prediction and groundtruth
#     labels = torch.unique(prediction)
#     gt_labels = torch.unique(groundtruth)
#     labels = labels[labels > 0]
#     gt_labels = gt_labels[gt_labels > 0]
#     for label in labels:  # for each predicted label
#       iou = torch.tensor(0.0)
#       best_gt_label = -1
#       for gt_label in gt_labels:
#         # compute iou with all instance gt labels and store the best
#         intersection = (((prediction == label) & (groundtruth == gt_label)).sum()).float()
#         union = ((prediction == label).sum() + (groundtruth == gt_label).sum() - intersection).float()
#         iou_tmp = intersection / union
#         if iou_tmp > iou:
#           iou = iou_tmp
#           best_gt_label = gt_label
#       # if the best iou is above 0.5, store the match pred_label-gt_label-iou
#       if iou > 0.5:
#         result = {
#             "pred_label": label.item(),
#             "gt_label": best_gt_label.item(),
#             "iou": iou.item(),
#         }
#         class_matches.append(result)
#         numerator_iou += iou
#       # else, the predicted label remains unmatched
#       else:
#         unmatched_class_predictions.append(label.item())

#     true_positives = len(class_matches)  # TP = number of matches
#     false_positives = len(unmatched_class_predictions)  # FP = number of unmatched predictions
#     false_negatives = len(gt_labels) - len(class_matches)  # FN = number of unmatched gt labels
#     panoptic_quality_one_class = numerator_iou / (true_positives + 0.5 * false_positives + 0.5 * false_negatives)
#     if true_positives == 0:
#       sq = 0.0
#       rq = 0.0
#     else:
#       sq = numerator_iou / true_positives
#       rq = true_positives / (true_positives + 0.5 * false_positives + 0.5 * false_negatives)

#     return panoptic_quality_one_class, sq, rq

#   def compute_pq(self, semantic_pred, semantic_gt, instance_pred, instance_gt):
#     # take the semantic labels (except 0 = void label)
#     semantic_gt_labels = torch.unique(semantic_gt).int()
#     semantic_gt_labels = semantic_gt_labels[semantic_gt_labels > 0].int()
#     for gt_label in semantic_gt_labels:
#       # take pred and gt where they have sem_class = label
#       semantic_tmp = (semantic_pred == gt_label).int()
#       semantic_gt_tmp = (semantic_gt == gt_label).int()

#       # take pred and gt instances corresponding to the considered semantic class
#       instance_tmp = (instance_pred * semantic_tmp).int()
#       instance_gt_tmp = (instance_gt * semantic_gt_tmp).int()

#       pq, sq, rq = self.compute_pq_single_class(instance_tmp, instance_gt_tmp)

#       # update: count = number of images that contains the class, pq is the overall panoptic quality of the images
#       if gt_label.item() not in self.panoptic_qualities.keys():
#         self.panoptic_qualities[gt_label.item()] = {}
#         self.panoptic_qualities[gt_label.item()]["count"] = 1
#         self.panoptic_qualities[gt_label.item()]["pq"] = pq.item()
#         self.panoptic_qualities[gt_label.item()]["sq"] = sq
#         self.panoptic_qualities[gt_label.item()]["rq"] = rq
#       else:
#         assert False  # "Call this function with batch size 1"
#         # self.panoptic_qualities[gt_label.item()]["count"] += 1
#         # self.panoptic_qualities[gt_label.item()]["pq"] = (
#         #     (
#         #         self.panoptic_qualities[gt_label.item()]["pq"]
#         #         * (self.panoptic_qualities[gt_label.item()]["count"] - 1)
#         #         + pq
#         #     )
#         #     / self.panoptic_qualities[gt_label.item()]["count"]
#         # ).item()

#     # panoptic_quality = self.average_pq(self.panoptic_qualities)
#     pdb.set_trace()
#     return (
#         [val["pq"] for key, val in self.panoptic_qualities.items()],
#         [val["sq"] for key, val in self.panoptic_qualities.items()],
#         [val["rq"] for key, val in self.panoptic_qualities.items()],
#     )

#   def average_pq(self, panoptic_qualities):
#     # overall panoptic quality averaged from all classes
#     pq = 0.0
#     n = 0
#     for item in panoptic_qualities.keys():
#       pq += panoptic_qualities[item]["pq"]
#       n += 1
#     if n == 0:
#       return -1
#     pq /= n
#     return pq
